'use strict';

exports.__esModule = true;
exports.startsBefore = startsBefore;
exports.positionFromDate = positionFromDate;
exports.default = getStyledEvents;

var _accessors = require('./accessors');

var _dates = require('./dates');

var _dates2 = _interopRequireDefault(_dates);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var checkIsMultidayStart = function checkIsMultidayStart(end, max) {
  return end > max;
};

var checkIsMultidayEnd = function checkIsMultidayEnd(start, min) {
  return start < min;
};

function startsBefore(date, min) {
  return _dates2.default.lt(_dates2.default.merge(min, date), min, 'minutes');
}

function positionFromDate(date, min, total) {
  if (startsBefore(date, min)) return 0;

  var diff = _dates2.default.diff(min, _dates2.default.merge(min, date), 'minutes');
  return Math.min(diff, total);
}

/**
 * Events will be sorted primarily according to earliest start time.
 * If two events start at the same time, the one with the longest duration will
 * be placed first.
 */
var sort = function sort(events, _ref) {
  var startAccessor = _ref.startAccessor,
      endAccessor = _ref.endAccessor;
  return events.sort(function (a, b) {
    var startA = +(0, _accessors.accessor)(a, startAccessor);
    var startB = +(0, _accessors.accessor)(b, startAccessor);

    if (startA === startB) {
      return +(0, _accessors.accessor)(b, endAccessor) - +(0, _accessors.accessor)(a, endAccessor);
    }

    return startA - startB;
  });
};

var getSlot = function getSlot(event, accessor, min, totalMin) {
  return event && positionFromDate((0, _accessors.accessor)(event, accessor), min, totalMin);
};

/**
 * Two events are considered siblings if the difference between their
 * start time is less than to 30 minutes.
 */
var isSibling = function isSibling(idx1, idx2, _ref2) {
  var events = _ref2.events,
      startAccessor = _ref2.startAccessor,
      min = _ref2.min,
      totalMin = _ref2.totalMin;

  var event1 = events[idx1];
  var event2 = events[idx2];

  if (!event1 || !event2) return false;

  var isSameDate = _dates2.default.sameDate(event1[startAccessor], event2[startAccessor]);

  if (!isSameDate) return false;

  var start1 = getSlot(event1, startAccessor, min, totalMin);
  var start2 = getSlot(event2, startAccessor, min, totalMin);

  return Math.abs(start1 - start2) < 30;
};

/**
 * An event is considered a child of another event if its start time is
 * more than 30 minutes later than the other event's start time,
 * but before its end time.
 */
var isChild = function isChild(parentIdx, childIdx, _ref3) {
  var events = _ref3.events,
      startAccessor = _ref3.startAccessor,
      endAccessor = _ref3.endAccessor,
      min = _ref3.min,
      totalMin = _ref3.totalMin,
      max = _ref3.max;

  if (isSibling(parentIdx, childIdx, { events: events, startAccessor: startAccessor, endAccessor: endAccessor, min: min, totalMin: totalMin })) return false;

  var parent = events[parentIdx];
  var child = events[childIdx];

  var isMultidayEnd = parent && checkIsMultidayEnd(parent[startAccessor], min);
  var isMultiday = parent && !isMultidayEnd && !_dates2.default.sameDate(parent[startAccessor], parent[endAccessor]);

  var parentEnd = getSlot(parent, endAccessor, min, totalMin);

  if (isMultiday) {
    var _getSlot;

    parentEnd = getSlot((_getSlot = {}, _getSlot[endAccessor] = max, _getSlot), endAccessor, min, totalMin);
  }

  var childStart = getSlot(child, startAccessor, min, totalMin);

  return parentEnd > childStart;
};

/**
 * Given an event index, siblings directly following it will be found and
 * returned as an array of indexes.
 */
var getSiblings = function getSiblings(idx, _ref4) {
  var events = _ref4.events,
      startAccessor = _ref4.startAccessor,
      endAccessor = _ref4.endAccessor,
      min = _ref4.min,
      totalMin = _ref4.totalMin;

  var nextIdx = idx;
  var siblings = [];

  while (isSibling(idx, ++nextIdx, { events: events, startAccessor: startAccessor, endAccessor: endAccessor, min: min, totalMin: totalMin })) {
    siblings.push(nextIdx);
  }

  return siblings;
};

/**
 * Given an event index, and a start search position, all child events to that
 * event will be found and placed into groups of siblings.
 * The return value is an array of child group arrays, as well as the largest
 * size of the child groups.
 */
var getChildGroups = function getChildGroups(idx, nextIdx, _ref5) {
  var events = _ref5.events,
      startAccessor = _ref5.startAccessor,
      endAccessor = _ref5.endAccessor,
      min = _ref5.min,
      totalMin = _ref5.totalMin,
      max = _ref5.max;

  var groups = [];
  var nbrOfColumns = 0;

  while (isChild(idx, nextIdx, { events: events, startAccessor: startAccessor, endAccessor: endAccessor, min: min, totalMin: totalMin, max: max })) {
    var childGroup = [nextIdx];
    var siblingIdx = nextIdx;

    while (isSibling(nextIdx, ++siblingIdx, { events: events, startAccessor: startAccessor, endAccessor: endAccessor, min: min, totalMin: totalMin })) {
      childGroup.push(siblingIdx);
    }

    nbrOfColumns = Math.max(nbrOfColumns, childGroup.length);
    groups.push(childGroup);
    nextIdx = siblingIdx;
  }

  return { childGroups: groups, nbrOfChildColumns: nbrOfColumns };
};

var getIfLastSlotOfTheDayEnd = function getIfLastSlotOfTheDayEnd(end, min, timezone) {
  var localTZone = _dates2.default.getLocalTimezone();
  var localTZoneOffset = _dates2.default.getTimezoneOffset(end, localTZone);
  var calendarTZoneOffset = _dates2.default.getTimezoneOffset(end, timezone);
  var normalizedEnd = _dates2.default.getTZNormalizedEndDate(end, localTZoneOffset - calendarTZoneOffset);
  return +normalizedEnd === +min;
};

/**
 * Returns height and top offset, both in percentage, for an event at
 * the specified index.
 */
var getYStyles = function getYStyles(idx, _ref6) {
  var events = _ref6.events,
      startAccessor = _ref6.startAccessor,
      endAccessor = _ref6.endAccessor,
      min = _ref6.min,
      totalMin = _ref6.totalMin,
      max = _ref6.max,
      step = _ref6.step,
      timezone = _ref6.timezone;

  var ev = events[idx];
  var isLastSlotOfTheDayEnd = getIfLastSlotOfTheDayEnd(ev.end, min, timezone);
  var isMultidayStart = checkIsMultidayStart(ev[endAccessor], max);
  var isMultidayEnd = checkIsMultidayEnd(ev[startAccessor], min);

  var start = isMultidayEnd ? 0 : getSlot(ev, startAccessor, min, totalMin);

  var end = isMultidayStart ? totalMin : Math.max(getSlot(ev, endAccessor, min, totalMin), start + step);

  if (isLastSlotOfTheDayEnd) {
    end = isMultidayStart ? totalMin : Math.max(getSlot(ev, endAccessor, min, totalMin), start);
  }

  var top = start / totalMin * 100;
  var bottom = end / totalMin * 100;

  return {
    isMultidayStart: isMultidayStart,
    isMultidayEnd: isMultidayEnd,
    top: top,
    height: bottom - top,
    padding: isLastSlotOfTheDayEnd && 0
  };
};

var SMALL_EVENT_WIDTH = 30;

var isSmallEvent = function isSmallEvent(width) {
  return width < SMALL_EVENT_WIDTH;
};

var isFirstSibling = function isFirstSibling(siblingIdx) {
  return siblingIdx === 0;
};

var isOnlyNestedElement = function isOnlyNestedElement(group) {
  return group.length === 1;
};

var resetEvent = function resetEvent(event) {
  delete event.overlappingCount;
  delete event.groupNumber;
  delete event.groupStart;
  delete event.groupEnd;
};

/**
 * Takes an array of unsorted events, and returns a sorted array
 * containing the same events, but with an additional style property.
 * These styles will position the events similarly to Google Calendar.
 *
 * The algorithm will start by sorting the array, and then iterating over it.
 * Starting at the first event, each of its siblings and children, placed in
 * groups of siblings, will be found. Both are needed in order to calculate the
 * width of the first event. When the width is known, its siblings will be
 * given the same width, but with an incremental x-offset.
 *
 * Each group of children will be looking to move as far away from its original
 * parent as possible. A move can be made to one of the parent's siblings, if
 * that sibling is also a parent to the child group. This will make room for
 * more events.
 *
 * When a child group knows its parent, it looks at the space occupied by that
 * parent, and calculates the remaning available space and divides that among
 * each other.
 *
 * All widths and x-offsets are calculated without taking overlapping into
 * account. Overlapping is added in the end according to the OVERLAP_MULTIPLIER.
 * If that is set to 0, the events won't overlap or grow.
 *
 * When one of these rounds are finished, all events connected have been
 * traversed, so the cursor will be moved past all of them.
 */
function getStyledEvents(_ref7) {
  var unsortedEvents = _ref7.events,
      startAccessor = _ref7.startAccessor,
      endAccessor = _ref7.endAccessor,
      min = _ref7.min,
      totalMin = _ref7.totalMin,
      step = _ref7.step,
      max = _ref7.max,
      timezone = _ref7.timezone;

  var events = sort(unsortedEvents, { startAccessor: startAccessor, endAccessor: endAccessor });
  var helperArgs = { events: events, startAccessor: startAccessor, endAccessor: endAccessor, min: min, totalMin: totalMin, step: step, max: max, timezone: timezone };
  var styledEvents = [];
  var idx = 0;

  // One iteration will cover all connected events.

  var _loop = function _loop() {
    var siblings = getSiblings(idx, helperArgs);

    var _getChildGroups = getChildGroups(idx, idx + siblings.length + 1, helperArgs),
        childGroups = _getChildGroups.childGroups;
    // Set styles to top level events.


    [idx].concat(siblings).forEach(function (eventIdx, siblingIdx) {
      var width = void 0;
      var xOffset = void 0;
      var event = events[eventIdx];

      resetEvent(event);

      if (siblings.length) {
        // styles for top level side-by-side events

        var siblingsNumber = siblings.length + 1;
        event.overlappingCount = siblingsNumber;

        event.groupStart = siblingIdx === 0;
        event.groupEnd = siblingIdx === siblings.length;

        width = 100 / siblingsNumber;
        xOffset = isFirstSibling(siblingIdx) ? 0 : width * siblingIdx;
      } else {
        // styles for top level single event
        width = 100;
        xOffset = 0;
      }

      var _getYStyles = getYStyles(eventIdx, helperArgs),
          top = _getYStyles.top,
          height = _getYStyles.height,
          isMultidayStart = _getYStyles.isMultidayStart,
          isMultidayEnd = _getYStyles.isMultidayEnd,
          padding = _getYStyles.padding;

      event.smallEvent = isSmallEvent(width);
      event.isMultidayStart = isMultidayStart;
      event.isMultidayEnd = isMultidayEnd;

      styledEvents[eventIdx] = {
        event: event,
        style: {
          top: top,
          height: height,
          width: width,
          xOffset: xOffset,
          padding: padding
        }
      };
    });

    childGroups.forEach(function (group, groupIndex) {
      var parentIdx = idx;
      var siblingIdx = 0;

      // Move child group to sibling if possible, since this will makes
      // room for more events.
      while (isChild(siblings[siblingIdx], group[0], helperArgs)) {
        parentIdx = siblings[siblingIdx];
        siblingIdx++;
      }

      // Set styles to child events.
      group.forEach(function (eventIdx, i) {
        var parentStyle = styledEvents[parentIdx].style;
        var parentWidth = parentStyle.width,
            parentXOffset = parentStyle.xOffset;
        // let columns = Math.min(group.length, nbrOfColumns);

        var event = events[eventIdx];

        resetEvent(event);

        var offset = 3;
        var nestedGroupOffset = 3;
        var groupNumber = groupIndex + 1;
        var overlappingCount = group.length;
        var width = void 0;
        var xOffset = void 0;

        // styles for not overlapping nested elements
        if (isOnlyNestedElement(group)) {
          xOffset = offset * groupNumber;
          width = 100 - xOffset;
        } else {
          var groupOffset = nestedGroupOffset * groupNumber;
          width = (parentWidth - offset * groupIndex) / overlappingCount - offset / overlappingCount;
          var childOffsetInGroup = width * i;
          xOffset = parentXOffset + childOffsetInGroup + groupOffset;
          event.overlappingCount = overlappingCount;
          event.groupNumber = groupNumber;
          event.groupStart = i === 0;
          event.groupEnd = i === group.length - 1;
        }

        var _getYStyles2 = getYStyles(eventIdx, helperArgs),
            top = _getYStyles2.top,
            height = _getYStyles2.height,
            padding = _getYStyles2.padding;

        event.smallEvent = isSmallEvent(width);

        styledEvents[eventIdx] = {
          event: event,
          style: {
            top: top,
            height: height,
            width: width,
            xOffset: xOffset,
            padding: padding
          }
        };
      });
    });

    // Move past all events we just went through
    idx += 1 + siblings.length + childGroups.reduce(function (total, group) {
      return total + group.length;
    }, 0);
  };

  while (idx < events.length) {
    _loop();
  }

  return styledEvents;
}